@@ -1,4 +1,4 @@
-/* Copyright 2002-2004 Elliotte Rusty Harold
+/* Copyright 2002-2006, 2009, 2010 Elliotte Rusty Harold
    
    This library is free software; you can redistribute it and/or modify
    it under the terms of version 2.1 of the GNU Lesser General Public 
@@ -53,7 +53,7 @@ import org.apache.xerces.impl.Version;
  * </p>
  * 
  * @author Elliotte Rusty Harold
- * @version 1.0
+ * @version 1.2.5
  * 
  */
 public class Builder {
@@ -67,9 +67,16 @@ public class Builder {
     static {  
 
         try {
-            String versionString = Version.getVersion();
-            versionString = versionString.substring(9, 12);
-            xercesVersion = Double.valueOf(versionString).doubleValue();
+            String x = Version.getVersion();
+            String versionString = x.substring(9);
+            int firstPeriod = versionString.indexOf(".");
+            int secondPeriod = versionString.lastIndexOf(".");
+            String major = versionString.substring(0, firstPeriod);
+            String minor = versionString.substring(firstPeriod+1, secondPeriod);
+            if (Integer.parseInt(minor) < 10 && Integer.parseInt(major) < 3) {
+                xercesVersion = Double.parseDouble(x.substring(9,12));
+            }
+            // else it's 2.6 or later which is all we really need to know
         }
         catch (Exception ex) {
             // The version string format changed so presumably it's
@@ -166,6 +173,8 @@ public class Builder {
         "nu.xom.XML1_0Parser",
         "nu.xom.JDK15XML1_0Parser",
         "org.apache.xerces.parsers.SAXParser",
+        "org.apache.xerces.jaxp.SAXParserImpl$JAXPSAXParser", // xerces-2.9.x
+        "com.sun.org.apache.xerces.internal.jaxp.SAXParserImpl$JAXPSAXParser", // JDK 1.6
         "com.sun.org.apache.xerces.internal.parsers.SAXParser",
         "gnu.xml.aelfred2.XmlReader",
         "org.apache.crimson.parser.XMLReaderImpl",
@@ -179,7 +188,7 @@ public class Builder {
     };
 
     
-    private static XMLReader findParser(boolean validate) {
+    static XMLReader findParser(boolean validate) {
         
         // first look for Xerces; we only trust Xerces if
         // we set it up; and we need to configure it specially
@@ -252,14 +261,22 @@ public class Builder {
     private static void setupParser(XMLReader parser, boolean validate)
       throws SAXNotRecognizedException, SAXNotSupportedException {
         
+        // General configuration for all parsers
+        parser.setFeature(
+                "http://xml.org/sax/features/namespace-prefixes", true);
+        parser.setFeature(
+                "http://xml.org/sax/features/namespaces", true);
+        
+        // Parser specific configuration
         XMLReader baseParser = parser;
         while (baseParser instanceof XMLFilter) {
-            baseParser = ((XMLFilter) baseParser).getParent();
+             XMLReader parent = ((XMLFilter) baseParser).getParent();
+             if (parent == null) break;
+             baseParser = parent;
         }
+        
         String parserName = baseParser.getClass().getName();
         if (!validate) {
-            parser.setFeature(
-              "http://xml.org/sax/features/namespace-prefixes", true);
             if (parserName.equals(  // Crimson workaround
               "org.apache.crimson.parser.XMLReaderImpl")) {
                 parser.setErrorHandler(
@@ -277,9 +294,7 @@ public class Builder {
                 );
             }
         }
-        else {  
-            parser.setFeature(
-              "http://xml.org/sax/features/namespace-prefixes", true);
+        else {
             parser.setFeature(
               "http://xml.org/sax/features/validation", true);
             parser.setErrorHandler(new ValidityRequired());
@@ -298,7 +313,10 @@ public class Builder {
         if (parserName.equals("nu.xom.XML1_0Parser") 
          || parserName.equals("nu.xom.JDK15XML1_0Parser")
          || parserName.equals("org.apache.xerces.parsers.SAXParser")
-         || parserName.equals("com.sun.org.apache.xerces.internal.parsers.SAXParser")) {
+         || parserName.equals("com.sun.org.apache.xerces.internal.parsers.SAXParser")
+         || parserName.equals("org.apache.xerces.jaxp.SAXParserImpl$JAXPSAXParser") // xerces-2.9.x
+         || parserName.equals("com.sun.org.apache.xerces.internal.jaxp.SAXParserImpl$JAXPSAXParser")) // JDK 1.6
+        {
             try {
                 parser.setFeature(
                  "http://apache.org/xml/features/allow-java-encodings", true);
@@ -446,8 +464,7 @@ public class Builder {
         // In general, a filter may violate the constraints of XML 1.0.
         // However, I specifically trust Norm Walsh not to do that, so 
         // if his filters are being used we look at the parent instead.
-        if (parserName.equals("org.apache.xml.resolver.tools.ResolvingXMLReader")
-          || parserName.equals("org.apache.xml.resolver.tools.ResolvingXMLFilter")) {
+        if (parserName.equals("org.apache.xml.resolver.tools.ResolvingXMLFilter")) {
             XMLFilter filter = (XMLFilter) parser;
             parserName = filter.getParent().getClass().getName();
         }
@@ -506,6 +523,13 @@ public class Builder {
             parser.setProperty(
               "http://xml.org/sax/properties/declaration-handler", 
               handler);
+            // Due to Crimson bugs in misidentifying the internal and
+            // external DTD subsets, we only build the internal DTD 
+            // subset if there is no external DTD subset.
+            if (parser.getClass().getName().equals(
+              "org.apache.crimson.parser.XMLReaderImpl")) {
+                handler.usingCrimson = true;
+            }
         }
         catch (SAXException ex) {
             // This parser does not support declaration events.
@@ -533,8 +557,7 @@ public class Builder {
      * that takes a <code>java.io.File</code> object as an argument.
      * </p>
      * 
-     * @param systemID the URL (generally absolute) 
-     *     from which the document is read.
+     * @param systemID an absolute URL from which the document is read.
      *     The URL's scheme must be one supported by the Java VM. 
      * 
      * @return the parsed <code>Document</code>
@@ -567,12 +590,14 @@ public class Builder {
      * @throws ValidityException if a validity error is detected; 
      *     only thrown if the builder has been instructed to validate
      * @throws ParsingException  if a well-formedness error is detected
-     * @throws IOException  if an I/O error such as a broken 
-     *     socket prevents the document from being fully read.
+     * @throws IOException       if an I/O error such as a broken 
+     *     socket prevents the document from being fully read
+     * @throws NullPointerException  if <code>in</code> is null
      */
     public Document build(InputStream in) 
       throws ParsingException, ValidityException, IOException {
 
+        if (in == null) throw new NullPointerException("Null InputStream");
         InputSource source = new InputSource(in);
         return build(source);
         
@@ -586,7 +611,7 @@ public class Builder {
      * </p>
      * 
      * @param in the input stream from which the document is read
-     * @param baseURI the base URI for this document
+     * @param baseURI an absolute URI for this document; may be null
      * 
      * @return the parsed <code>Document</code>
      * 
@@ -599,9 +624,11 @@ public class Builder {
     public Document build(InputStream in, String baseURI) 
       throws ParsingException, ValidityException, IOException {
 
-        baseURI = canonicalizeURL(baseURI);
         InputSource source = new InputSource(in);
-        source.setSystemId(baseURI);
+        if (baseURI != null) {
+            baseURI = canonicalizeURL(baseURI);
+            source.setSystemId(baseURI);
+        }
         return build(source);
         
     }
@@ -610,16 +637,14 @@ public class Builder {
     // Nasty hack to make sure we get the right form
     // of file URLs on Windows
     private static String fileURLPrefix = "file://";
-    private static boolean isWindows = false;
     
     static {
         String os = System.getProperty("os.name", "Unix");
         // I could do System.setProperty("os.name" "Windows") to test 
-        // this, but I'd need to us ea frsh ClassLoader to rerun the
+        // this, but I'd need to use a fresh ClassLoader to rerun the
         // static initializer block.
         if (os.indexOf("Windows") >= 0) {
             fileURLPrefix = "file:/";
-            isWindows = true;
         }
     }
 
@@ -953,7 +978,7 @@ public class Builder {
                             i++;
                             try {
                                 char low = absolute.charAt(i);
-                                String character = String.valueOf(c)+String.valueOf(low);
+                                String character = String.valueOf(c)+low;
                                 byte[] data = character.getBytes("UTF8");
                                 // Always exactly 4 bytes, unless the encoder is buggy
                                 for (int j=0; j < 4; j++) {
@@ -964,14 +989,14 @@ public class Builder {
                             }
                             catch (IndexOutOfBoundsException ex) {
                                 // file name contains a high half and not a low half
-                                url = new StringBuffer();
+                                url = new StringBuffer(0);
                                 break;
                             }
                         }
                         else {
                             // low half not preceded by high half
                             // Can't create a base URI
-                            url = new StringBuffer();
+                            url = new StringBuffer(0);
                             break;
                         }
                 }
@@ -1008,6 +1033,7 @@ public class Builder {
     public Document build(Reader in) 
       throws ParsingException, ValidityException, IOException {
 
+        if (in == null) throw new NullPointerException("Attempted to build from null reader");
         InputSource source = new InputSource(in);
         return build(source);
         
@@ -1022,7 +1048,7 @@ public class Builder {
      * 
      * @param in the reader from which the document 
      *     is read
-     * @param baseURI the base URI for this document
+     * @param baseURI the base URI for this document; may be null
      * 
      * @return the parsed <code>Document</code>
      * 
@@ -1035,10 +1061,12 @@ public class Builder {
      */
     public Document build(Reader in, String baseURI) 
       throws ParsingException, ValidityException, IOException {
-          
-        baseURI = canonicalizeURL(baseURI);
+         
         InputSource source = new InputSource(in);
-        source.setSystemId(baseURI);
+        if (baseURI != null) {
+            baseURI = canonicalizeURL(baseURI);
+            source.setSystemId(baseURI);
+        }
         return build(source);
         
     }
@@ -1049,9 +1077,8 @@ public class Builder {
      * Reads the document from the contents of a string.
      * </p>
      * 
-     * @param document the string that contains 
-     *      the XML document. 
-     * @param baseURI the base URI for this document
+     * @param document the string that contains the XML document
+     * @param baseURI the base URI for this document; may be null
      * 
      * @return  the parsed <code>Document</code>
      * 
@@ -1072,17 +1099,40 @@ public class Builder {
     
     // needed to work around a bug in Xerces and Crimson
     // for URLs with no trailing slashes (no path part) 
-    // such as http://www.cafeconleche.org
+    // such as http://www.cafeconleche.org.
+    // Also needed to work around a VM bug involving file URLs such as
+    // file:///tmp/nosuchdirectory/../foo.xml
+    // where "nosuchdirectory" does not exist.
     private String canonicalizeURL(String uri) {
         
         try {
             URL u = new URL(uri);
             String path = u.getFile();
+            String scheme = u.getProtocol();
+            String authority = u.getHost();
+            String query = u.getQuery();
+            int port = u.getPort();
+            // fragment ID not needed
             if (path == null || path.length() == 0 
+              // We handle here the case where we have a URL such as 
+              // http://www.cafeaulait.org with no trailing slash.
+              // Java's URL class assigns the path "/" to this case
+              // but does not change the URL. 
               || ("/".equals(path) && !(uri.endsWith("/")))) {
-                uri += '/';
+                path += '/';
             }
-            return uri;
+            // If this proves to be a hot spot we could probably take this path
+            // only if the scheme is file; not in the more common case where
+            // it's http
+            path = URIUtil.removeDotSegments(path);
+            StringBuffer canonicalForm = new StringBuffer(uri.length()); 
+            canonicalForm.append(scheme);
+            canonicalForm.append("://");
+            if (authority != null) canonicalForm.append(authority); 
+            if (port >= 0) canonicalForm.append(":" + port);
+            canonicalForm.append(path);
+            if (query != null) canonicalForm.append("?" + query);
+            return canonicalForm.toString();
         }
         catch (MalformedURLException ex) {
             return uri;
@@ -1095,8 +1145,7 @@ public class Builder {
      * Reads the document from a SAX <code>InputSource</code>.
      * </p>
      * 
-     * @param in the input source from 
-     *     which the document is read. 
+     * @param in the input source from which the document is read
      * 
      * @return the parsed <code>Document</code>
      * 
@@ -1110,8 +1159,11 @@ public class Builder {
     private Document build(InputSource in) 
       throws ParsingException, ValidityException, IOException {
 
+        XOMHandler handler = (XOMHandler) parser.getContentHandler();
+        Document result = null;
         try {
             parser.parse(in);
+            result = handler.getDocument();
         }
         catch (SAXParseException ex) {
             ParsingException pex = new ParsingException(
@@ -1119,7 +1171,7 @@ public class Builder {
                 ex.getSystemId(),
                 ex.getLineNumber(),
                 ex.getColumnNumber(),
-                ex.getException());
+                ex);
             throw pex;
         }
         catch (SAXException ex) {
@@ -1162,14 +1214,23 @@ public class Builder {
                 throw ex;
             }
         }
+        finally {
+            handler.freeMemory();
+        }
         
-        XOMHandler handler = (XOMHandler) parser.getContentHandler();
-        ErrorHandler errorHandler = parser.getErrorHandler();
-        Document result = handler.getDocument();
-        if (result != null && "".equals(result.getBaseURI())) {
+        if (result == null) {
+            ParsingException ex = new ParsingException(
+              "Parser did not build document", 
+              in.getSystemId(), -1, -1
+            );
+            throw ex;
+        }
+        
+        if ("".equals(result.getBaseURI())) {
             result.setBaseURI(in.getSystemId());
         }
         
+        ErrorHandler errorHandler = parser.getErrorHandler();
         if (errorHandler instanceof ValidityRequired) {
             ValidityRequired validityHandler 
               = (ValidityRequired) errorHandler;
@@ -1231,8 +1292,14 @@ public class Builder {
         }
       
         public void error(SAXParseException exception) 
-          throws SAXParseException { 
-            throw exception;            
+          throws SAXParseException {
+            
+            if (exception.getMessage().equals("Illegal Namespace prefix: \"xml\".")) {
+                return;
+            }
+            
+            throw exception;
+            
         }
       
         public void fatalError(SAXParseException exception) 
@@ -1246,8 +1313,8 @@ public class Builder {
     // I added this because XIncluder needed it.
     /**
      * <p>
-     * Returns this builder's <code>NodeFactory</code>. It may return
-     * null if a factory was not supplied when the builder was created. XXX
+     * Returns this builder's <code>NodeFactory</code>. It returns
+     * null if a factory was not supplied when the builder was created.
      * </p>
      * 
      * @return the node factory that was specified in the constructor
